/*
 * ControlFrame.java
 *
 * Created on April 16, 2003, 7:46 PM
 */

/**
 *
 * @author  rod
 */

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import actor.*;
import geometry.*;

public class ControlFrame extends JFrame implements ActionListener {
    
    JTextField bk_p;
    JTextField bk_theta;
    JTextField bk_v;
    JTextField bk_omega;
    JTextField btheta_turn;
    JTextField bepsilon;
    JTextField brepulsionForce;
    JTextArea actorInfo;
    
    /** Creates a new instance of ControlFrame */
    public ControlFrame() {
        //super();
        Rectangle virtualBounds = new Rectangle();
        GraphicsEnvironment ge = GraphicsEnvironment.
        getLocalGraphicsEnvironment();
        GraphicsDevice[] gs =
        ge.getScreenDevices();
        for (int j = 0; j < gs.length; j++) {
            GraphicsDevice gd = gs[j];
            GraphicsConfiguration[] gc =
            gd.getConfigurations();
            for (int i=0; i < gc.length; i++) {
                virtualBounds =
                virtualBounds.union(gc[i].getBounds());
            }
        }
        this.setLocation((int)(virtualBounds.getMaxX()-2*virtualBounds.getWidth()/3), (int)(virtualBounds.getHeight()/3));
        initFrame();
        pack();
        this.setResizable(false);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        show();
    }
    
    void initFrame(){
        setTitle("Control");
        Container contentPane = getContentPane();
        JPanel mainPanel = new JPanel();
        mainPanel.setLayout(new BorderLayout());
        
        /*CONSTANTS*/
        JPanel constantPanel = new JPanel();
        JLabel bk_p_label = new JLabel("<HTML><i>k<sub>p</sub></i></HTML>");
        bk_p = new JTextField(""+Universe.k_p, 4);
        bk_p.addActionListener(this);
        bk_p.setActionCommand("Constant:k_p");
        constantPanel.add(bk_p_label);
        constantPanel.add(bk_p);
        JLabel bk_v_label = new JLabel("<HTML><i>k<sub>v</sub></i></HTML>");
        bk_v = new JTextField(""+Universe.k_v, 4);
        bk_v.addActionListener(this);
        bk_v.setActionCommand("Constant:k_v");
        constantPanel.add(bk_v_label);
        constantPanel.add(bk_v);
        JLabel bepsilon_label = new JLabel("<HTML><i>&#949</i></HTML>");
        bepsilon = new JTextField(""+Universe.epsilon, 4);
        bepsilon.addActionListener(this);
        bepsilon.setActionCommand("Constant:epsilon");
        constantPanel.add(bepsilon_label);
        constantPanel.add(bepsilon);
        JLabel brepulsionForce_label = new JLabel("<HTML><i>F<sub>r</sub></i></HTML>");
        brepulsionForce = new JTextField(""+Universe.repulsionForce, 4);
        brepulsionForce.addActionListener(this);
        brepulsionForce.setActionCommand("Constant:repulsionForce force");
        constantPanel.add(brepulsionForce_label);
        constantPanel.add(brepulsionForce);
        JButton apply = new JButton("Apply");
        apply.setActionCommand("Constant");
        apply.addActionListener(this);
        constantPanel.add(apply);
        mainPanel.add(constantPanel);
        contentPane.add(mainPanel, BorderLayout.NORTH);
        
        /*ADD ACTORS*/
        JPanel centrePanel = new JPanel();
        JButton bperson = new JButton("Add Person");
        bperson.setActionCommand("Person");
        bperson.addActionListener(this);
        JButton bsquare = new JButton("Add Square");
        bsquare.setActionCommand("Square");
        bsquare.addActionListener(this);
        JButton btriangle = new JButton("Add Triangle");
        btriangle.setActionCommand("Triangle");
        btriangle.addActionListener(this);
        JButton bgoal = new JButton("Add Goal");
        bgoal.setActionCommand("Goal");
        bgoal.addActionListener(this);
        centrePanel.add(bperson);
        centrePanel.add(bsquare);
        centrePanel.add(btriangle);
        centrePanel.add(bgoal);
        contentPane.add(centrePanel, BorderLayout.CENTER);
        
        /*ACTOR INFO*/
        JPanel bottom = new JPanel();
        bottom.setBorder(BorderFactory.createEtchedBorder());
        actorInfo = new JTextArea("Information about actors on the screen will appear here", 5, 30);
        JScrollPane jsp = new JScrollPane(actorInfo);
        bottom.add(jsp);
        contentPane.add(bottom, BorderLayout.SOUTH);
    }
    
    public void setDescription(String desc){
        actorInfo.setText(desc);
    }
    
    private JPanel initialCentrePanel(JPanel cp){
        return cp;
    }
    
    static geometry.Point offset = new geometry.Point(0,0);
    static geometry.Point offsetIncrement = new geometry.Point(1.5, 1.5);

    private geometry.Point offset(){
        geometry.Point ret = offset;
        offset=offset.add(offsetIncrement);
        return ret;
    }
    
    public void actionPerformed(java.awt.event.ActionEvent actionEvent) {
        String cmd = actionEvent.getActionCommand();
        if (cmd.startsWith("Constant")){
            Universe.k_p = (new Double(bk_p.getText())).doubleValue();
            bk_p.setText((new Double(Universe.k_p)).toString());
            Universe.k_v = (new Double(bk_v.getText())).doubleValue();
            bk_v.setText((new Double(Universe.k_v)).toString());
            Universe.epsilon = (new Double(bepsilon.getText())).doubleValue();
            bepsilon.setText((new Double(Universe.epsilon)).toString());
            Universe.repulsionForce = (new Double(brepulsionForce.getText())).doubleValue();
            brepulsionForce.setText((new Double(Universe.repulsionForce)).toString());
        }
        if (cmd.equals("Person")){
            Person p = new Person(Universe.centre().add(offset()), new Vec(0,3), new Vec(), Color.blue, 1.0);
            Universe.add(p);
            System.out.println("New Person: "+p.name());
        }
        if (cmd.equals("Square")){
            Square s = new Square(Universe.centre().add(offset()), Color.black);
            Universe.add(s);
            System.out.println("New Obstacle (Square): "+s.name());
        }
        if (cmd.equals("Triangle")){
            Triangle s = new Triangle(Universe.centre().add(offset()), Color.orange);
            Universe.add(s);
            System.out.println("New Obstacle (Triangle). :"+s.name());
        }
        if (cmd.equals("Goal")){
            Goal g = new Goal(Color.red, 0.5, Universe.centre().add(offset()));
            Universe.add(g);
            System.out.println("New Goal: "+g.name());
        }
    }
}
