/*
 * SimulationFrame.java
 *
 * Created on April 16, 2003, 7:48 PM
 */

/**
 *
 * @author  rod
 */

import geometry.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import actor.*;

public class SimulationFrame extends JFrame implements MouseListener, MouseMotionListener, MouseWheelListener, ComponentListener, ActionListener {
    
    static double scale = 10.0;
    Universe universe;
    simulationPanel playPen;
    Graphics gr;
    JButton pauseUnpauseButton;
    
    /** Creates a new instance of SimulationFrame */
    public SimulationFrame() {
        gr = getGraphics();
        playPen = new simulationPanel();
        playPen.addMouseListener(this);
        playPen.addMouseMotionListener(this);
        playPen.addComponentListener(this);
        playPen.addMouseWheelListener(this);
        playPen.setSize(400,600);
        playPen.setBackground(Color.white);
        universe = new Universe(playPen, scale, this);
        initFrame();
        Rectangle virtualBounds = new Rectangle();
        GraphicsEnvironment ge = GraphicsEnvironment.
        getLocalGraphicsEnvironment();
        GraphicsDevice[] gs = ge.getScreenDevices();
        for (int j = 0; j < gs.length; j++) {
            GraphicsDevice gd = gs[j];
            GraphicsConfiguration[] gc =
            gd.getConfigurations();
            for (int i=0; i < gc.length; i++) {
                virtualBounds =
                virtualBounds.union(gc[i].getBounds());
            }
        }
        this.setLocation((int)(2*virtualBounds.getWidth()/3+virtualBounds.getMinX()), (int)(virtualBounds.getHeight()/3));
        setSize(playPen.getSize());
        initFrame();
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        show();
        //Thread t = new Thread(Universe.moveAll);
        //t.start();
        new javax.swing.Timer(50, Universe.moveAll).start();
    }
    
    private void initFrame(){
        setTitle("Simulation");
        Container contentPane = getContentPane();
        JPanel mainPanel = new JPanel();
        //mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
        mainPanel.setLayout(new BorderLayout());
        //mainPanel.setLayout(new GridLayout(2,1));
        JPanel buttonPanel = new JPanel();
        pauseUnpauseButton = new JButton("Start");
        pauseUnpauseButton.addActionListener(this);
        JButton exitButton = new JButton("Exit");
        exitButton.addActionListener(this);
        buttonPanel.add(pauseUnpauseButton);
        buttonPanel.add(exitButton);
        mainPanel.add(buttonPanel, BorderLayout.NORTH);
        playPen.setBorder(BorderFactory.createEtchedBorder());
        mainPanel.add(playPen, BorderLayout.CENTER);
        contentPane.add(mainPanel);
        Universe.updatePanel();
    }
    
    public void addActor(Actor a){
        Universe.add(a);
    }
    
    
    /*####################EVENTS##########################*/
    public void mouseClicked(java.awt.event.MouseEvent mouseEvent) {
        if (mouseEvent.getClickCount()==2){
            geometry.Point up = Universe.s2u(mouseEvent.getX(), mouseEvent.getY());
            Actor a = findActor(up);
            if (a != null){
                a.showControl();
            }
        }
    }
    
    public void mouseEntered(java.awt.event.MouseEvent mouseEvent) {
    }
    
    public void mouseExited(java.awt.event.MouseEvent mouseEvent) {
    }
    
    /*dragging stuff around*/
    Actor beingDragged = null;
    Person changeVel = null;
    geometry.Point startPosition = null;
    geometry.Point endPosition = null;
    geometry.Point diff = null;
    Actor beingResized = null;
    double initResizeDist = 0;
    boolean moving = false;
    boolean resizing = false;
    boolean goalsetting = false;
    Person gettingGoals;
    
    private void setResizeCursor(double angle){
        int cursor = 0;
        if (angle > 7*Math.PI/8 || angle < -7*Math.PI/8) cursor = Cursor.W_RESIZE_CURSOR;
        else if (angle < -5*Math.PI/8) cursor = Cursor.NW_RESIZE_CURSOR;
        else if (angle < -3*Math.PI/8) cursor = Cursor.S_RESIZE_CURSOR;
        else if (angle < -1*Math.PI/8) cursor = Cursor.NE_RESIZE_CURSOR;
        else if (angle < 1*Math.PI/8) cursor = Cursor.E_RESIZE_CURSOR;
        else if (angle < 3*Math.PI/8) cursor = Cursor.SE_RESIZE_CURSOR;
        else if (angle < 5*Math.PI/8) cursor = Cursor.N_RESIZE_CURSOR;
        else if (angle < 7*Math.PI/8) cursor = Cursor.SW_RESIZE_CURSOR;
        setCursor(Cursor.getPredefinedCursor(cursor));
    }
    
    public void mouseWheelMoved(MouseWheelEvent e){
        int zoom = -e.getWheelRotation();
        geometry.Point pre = Universe.centre();
        Universe.scale *= Math.pow(2,zoom);
        Universe.updatePanel();
        geometry.Point post = Universe.centre();
        Universe.moveOrigin(post.sub(pre));
        Universe.updatePanel();
        
        System.out.println("new scale "+Universe.scale+" px : 1 m");
    }
    
    public void mousePressed(java.awt.event.MouseEvent mouseEvent) {
        geometry.Point up = Universe.s2u(mouseEvent.getX(), mouseEvent.getY());
        Actor a = findActor(up);
        if(!mouseEvent.isShiftDown()){
            goalsetting = false;
            gettingGoals = null;
        }
        if (goalsetting == true && a != null){
            if (a.canBeGoal()){
                gettingGoals.pushGoal((GoalStackable)a, 1.0);
                System.out.println("Pushed goal "+a.name()+" on "+gettingGoals.name()+"'s stack.");
            }
        }
        if (a != null && a instanceof Person && mouseEvent.getButton() == MouseEvent.BUTTON1 && mouseEvent.isShiftDown()){
            goalsetting = true;
            gettingGoals = (Person)a;
            System.out.println("Setting multiple goals for "+gettingGoals.name());
        }
        if (mouseEvent.getButton() == MouseEvent.BUTTON1){
            moving = true;
            setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
            if (a == null){
                Person p = findVector(up);
                if (p == null){
                    // drag screen
                    startPosition = up;
                }else{
                    changeVel = p;
                    startPosition = p.position();//up.add(p.velocity());
                }
            }else{
                //move actor
                beingDragged = a;
                startPosition = a.position();
                endPosition = up;
                diff = up.sub(startPosition);
            }
            return;
        }else{ // assuming if it's not left, it's right...
            /*drag to resize*/
            if (a != null){
                resizing = true;
                beingResized = a;
                startPosition = a.position();
                endPosition = up;
                diff = up.sub(startPosition);
                double angle = Math.atan2(diff.yValue(), diff.xValue());
                setResizeCursor(angle);
            }
        }
    }
    
    public void mouseDragged(java.awt.event.MouseEvent mouseEvent){
        geometry.Point ep = Universe.s2u(mouseEvent.getX(), mouseEvent.getY());
        if (moving){
            if (beingDragged != null){
                if (beingDragged instanceof Obstacle){
                    ((Obstacle)beingDragged).setLocationRelative(ep.sub(endPosition));
                }else{
                    beingDragged.setPosition(ep.sub(diff));
                }
                beingDragged.draw(gr);
            }else if (changeVel != null){
                changeVel.setVelocity(new Vec(ep.sub(startPosition)));
                if (gr != null){
                    changeVel.velocity().draw(gr, changeVel.position(), Color.green);
                }
            }else{
                Universe.moveOrigin(ep.sub(startPosition));
            }
            endPosition = ep;
            return;
        }else if (resizing){ // resizing
            if (beingResized!=null){
                beingResized.resize(startPosition, ep);
                diff = ep.sub(startPosition);
                double angle = Math.atan2(diff.yValue(), diff.xValue());
                setResizeCursor(angle);
            }
        }
    }
    
    public void mouseMoved(java.awt.event.MouseEvent mouseEvent){
        geometry.Point up = Universe.s2u(mouseEvent.getX(), mouseEvent.getY());
        Actor a = findActor(up);
        if (a != null){
            setCursor(Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR));
            if (crowd.control != null)
                crowd.control.setDescription(a.description());
        }else{
            setCursor(Cursor.getDefaultCursor());
        }
    }
    
    public void mouseReleased(java.awt.event.MouseEvent mouseEvent) {
        geometry.Point up = Universe.s2u(mouseEvent.getX(), mouseEvent.getY());
        setCursor(Cursor.getDefaultCursor());
        if(moving){
            moving=false;
            if (beingDragged == null){
                if (changeVel != null){
                    changeVel = null;
                    return;
                }
                // drag screen
                if (startPosition == null) return;
                Universe.moveOrigin(up.sub(startPosition));
                startPosition = null;
            }else{
                beingDragged = null;
                startPosition = null;
                diff = null;
                for(int i = 0; i < Universe.actors.size(); i++){
                    Actor a = (Actor)(Universe.actors.elementAt(i));
                    if (a instanceof Person){
                        ((Person)a).removeSecondaryGoals();
                    }
                }
            }
        }else if (resizing){
            beingResized = null;
            resizing = false;
        }
    }
    
    public void componentHidden(java.awt.event.ComponentEvent componentEvent) {
    }
    
    public void componentMoved(java.awt.event.ComponentEvent componentEvent) {
    }
    
    public void componentResized(java.awt.event.ComponentEvent componentEvent) {
        if (universe != null) Universe.updatePanel();
    }
    
    public void componentShown(java.awt.event.ComponentEvent componentEvent) {
    }
    
    public void actionPerformed(ActionEvent evt){
        String c = evt.getActionCommand();
        if (c.equals("Exit")){
            System.exit(-1);
        }
        Universe.pauseUnpause();
        c=pauseUnpauseButton.getText();
        if (c.equals("Start")){
            pauseUnpauseButton.setText("Pause");
            return;
        }
        pauseUnpauseButton.setText("Start");
    }
    
    private Actor findActor(geometry.Point up){
        for (int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Person){
                double d = up.distanceTo(a.position());
                if (d <= ((Person)a).personalSpaceRadius()){
                    return a;
                }
            }else if (a instanceof Goal){
                double d = up.distanceTo(a.position());
                if (d <= ((Goal)a).radius()){
                    return a;
                }
            }else if (a instanceof Obstacle){
                if (((Obstacle)a).contains(up)){
                    return a;
                }
            }else
                System.out.println("What's a "+a.toString()+"?");
        }
        return null;
    }
    
    private Person findVector(geometry.Point up){
        for (int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Person){
                Person p = (Person)a;
                double d = up.distanceTo(p.position().add(p.velocity()));
                if (d <= (Universe.u2s(0.5))){
                    return p;
                }
            }
        }
        return null;
    }
    
    class simulationPanel extends JPanel{
        
        public simulationPanel(){
            super();
        }
        
        public void paintComponent(Graphics g){
            super.paintComponent(g);
            Universe.drawAll(g);
        }
    }
}
