/*
 * MovingActor.java
 *
 * Created on March 31, 2003, 9:54 PM
 */

package actor;

import geometry.*;
import java.awt.Graphics;
import javax.swing.JPanel;
/**
 *
 * @author  rod
 */
public abstract class MovingActor extends Actor {
    protected Vec velocity;
    protected Vec acceleration;
    private long lastTime;
    public double maxV;
    double maxA;
    
    /** Creates a new instance of MovingActor */
    public MovingActor(Point p, Vec initVel, Vec initAcc, double maxVel, double maxAcc) {
        super(p);
        velocity = initVel;
        acceleration = initAcc;
        lastTime = System.currentTimeMillis();
        maxV=maxVel;
        maxA=maxAcc;
    }
    
    public MovingActor(Point p) {
        super(p);
        velocity = new Vec();
        acceleration = new Vec();
        lastTime = System.currentTimeMillis();
        maxV=5;
        maxA=10;
    }
    
    public void start(){ // call when ready to start animation
        lastTime = System.currentTimeMillis();
    }
    
    public double speed(){
        return velocity.length();
    }
    
    public double acceleration(){
        return acceleration.length();
    }
    
    public Vec velocity(){
        return velocity;
    }
    
    public void setVelocity(Vec v){
        velocity = v;
    }
    
    public void setAcceleration(Vec acc){
        acceleration = acc;
        if (acceleration.length() > maxA){
            acceleration.setLength(maxA);
        }
    }
    
    public void update(){ //do movement
        if (Universe.paused){
            lastTime = System.currentTimeMillis();
            return;
        }
        /*if (acceleration.length() > maxA){
            acceleration.setLength(maxA);
        }*/
        long now = System.currentTimeMillis();
        double dt = (double)(now - lastTime)/1000.0;
        lastTime = now;
        velocity = velocity.add(acceleration.scalarmul(dt));
        Point ds = velocity.mul(dt);
        Point newPos = position().add(ds);
        for (int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Obstacle){
                Obstacle o = (Obstacle)a;
                if (o.contains(newPos)){
//                    System.out.print("["+name+"] Gosh, I'm in an object. Old velocity: "+velocity.toString());
                    Point cp = o.closestPoint(newPos);
                    Line cb = o.closestBorder(newPos);
                    velocity = velocity.reflect(cb);
                    ds = velocity.mul(dt);
                    newPos = position().add(ds);
//                    System.out.println(" New velocity: "+velocity.toString());
                }
            }
        }
        if (velocity.length() > maxV){
            velocity.setLength(maxV);
        }
        if (velocity.length() < Universe.epsilon){
            velocity.setLength(Universe.epsilon);
        }
        setPosition(newPos);
    }
    
    abstract public void move();
    
    abstract public String description();
    
    public abstract void showControl();
    
    public boolean canBeGoal(){
        return true;
    }
    
    public abstract JPanel controlPanel();
    
    public abstract void resize(Point start, Point end);
    
}
