/*
 * Obstacle.java
 *
 * Created on March 31, 2003, 6:06 PM
 */

package actor;

import geometry.*;
import java.util.Vector;
import java.awt.Polygon;
import java.awt.Graphics;
import java.awt.Color;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.Container;

/**
 *
 * @author  rod
 */
public abstract class Obstacle extends Actor implements ChangeListener, ActionListener{
    protected Vector border; // of Point
    private int cursor;
    protected Color color;
    static int obstNum = 0;
    
    /** Creates a new instance of Obstacle */
    public Obstacle(Point loc, Vector border, Color c){
        super(loc);
        this.border = border;
        setLocation(loc);
        color = c;
        cursor = 0;
        setName("Obstacle "+(++obstNum));
    }
    
    public boolean canBeGoal(){
        return false;
    }
    
    public boolean contains(Point p){ // is the point inside the object?
        /* if a ray from the point crosses the boundary an odd number of times, the point is inside
         * not using the Polygon.inside() methods because (a) I already did all this geometry work,
         * (b) it doesn't seem all that accurate, and (c) now that I've done all this stuff, I would
         * love to see if it works.
         */
        
        boolean inside = false;
        Line ray = new Line(p, Math.random()*10.0-5.0); // any old ray from the point
        for(int i = 0; i < border.size(); i++){
            if (ray.intersectRay(borderLine(i))!=null){
                inside = !inside;
            }
        }
        return inside;
    }
    
    public Point blocks(Line axis){ // some part of the object blocks the sight line
        // axis is defined as a ray viewpoint -> goalpoint
        Point res = null;
        Point sp = axis.startPoint();
        for(int i = 0; i < border.size(); i++){
            Line l = borderLine(i);
            Point in = axis.intersectRay(l);
            if (in != null){
                if ((res == null) || (in.distanceTo2(sp) < res.distanceTo2(sp)))
                    res=in;
            }
        }
        return res;
    }
    
    private Line borderLine(int index){
        if (index == border.size() - 1){
            return new Line((Point)(border.elementAt(index)), (Point)(border.elementAt(0)));
        }
        return new Line((Point)(border.elementAt(index)), (Point)(border.elementAt(index+1)));
    }
    
    public Point closestCornerPoint(Point p){ // point from border (i.e. a corner) closest to parameter
        //just go through them all, keep smallest
        double mindistance = p.distanceTo2((Point)(border.elementAt(0)));
        Point res = (Point)(border.elementAt(0));
        for(int i = 1; i < border.size(); i++){
            Point q = (Point)(border.elementAt(i));
            double dist = p.distanceTo2(q);
            if(dist < mindistance){
                mindistance = dist;
                res = q;
            }
        }
        return res;
    }
    
    public int closestCornerIndex(Point p){ // point from border (i.e. a corner) closest to parameter
        //just go through them all, keep smallest
        double mindistance = p.distanceTo2((Point)(border.elementAt(0)));
        int res = 0;
        for(int i = 1; i < border.size(); i++){
            Point q = (Point)(border.elementAt(i));
            double dist = p.distanceTo2(q);
            if(dist < mindistance){
                mindistance = dist;
                res = i;
            }
        }
        return res;
    }
    
    public Line closestBorder(Point p){
        int corner = closestCornerIndex(p);
        int left = corner - 1;
        if (left == -1){
            left = border.size()-1;
        }
        int right = corner + 1;
        if (right == border.size()){
            right = 0;
        }
        Line lb = new Line((Point)(border.elementAt(left)), (Point)(border.elementAt(corner)));
        Line rb = new Line((Point)(border.elementAt(right)), (Point)(border.elementAt(corner)));
        Point lp = lb.project(p);
        Point rp = rb.project(p);
        double ld = 0;
        double rd = 0;
        if (lb.contains(lp)){
            ld = p.distanceTo2(lp);
        }else{
            lp = (Point)(border.elementAt(corner));
            ld = p.distanceTo2(lp);
        }
        if (rb.contains(rp)){
            rd = p.distanceTo2(rp);
        }else{
            rp = (Point)(border.elementAt(corner));
            rd = p.distanceTo2(rp);
        }
        
        if (ld < rd){
            return lb;
        }
        return rb;
    }
    
    public Point closestPoint(Point p){ // closest point (for real) -- on border
        /* two possibilities here. Get closest corner, then find distance from point to intersection of perp defined
         * by each border line going left or right from that corner.
         */
        
        int corner = closestCornerIndex(p);
        int left = corner - 1;
        if (left == -1){
            left = border.size()-1;
        }
        int right = corner + 1;
        if (right == border.size()){
            right = 0;
        }
        Line lb = new Line((Point)(border.elementAt(left)), (Point)(border.elementAt(corner)));
        Line rb = new Line((Point)(border.elementAt(right)), (Point)(border.elementAt(corner)));
        Point lp = lb.project(p);
        Point rp = rb.project(p);
        double ld = 0;
        double rd = 0;
        if (lb.contains(lp)){
            ld = p.distanceTo2(lp);
        }else{
            lp = (Point)(border.elementAt(corner));
            ld = p.distanceTo2(lp);
        }
        if (rb.contains(rp)){
            rd = p.distanceTo2(rp);
        }else{
            rp = (Point)(border.elementAt(corner));
            rd = p.distanceTo2(rp);
        }
        
        if (ld < rd){
            return lp;
        }
        return rp;
    }
    
    public double distanceToClosestPoint(Point p){
        Point closest = closestPoint(p);
        return p.distanceTo(closest);
    }
    
    public Point leftEdge(Line axis){
        double spy = axis.startPoint().yValue();
        double spx = axis.startPoint().xValue();
        double axisAngle = Math.atan2(axis.endPoint().yValue() - spy, axis.endPoint().xValue() - spx);
        int i = 0;
        Point res = null;
        double maxangle = 2.0*Math.PI;
        while(i < border.size()){
            Point test = (Point)(border.elementAt(i));
            Line rl = new Line(axis.startPoint(), test);
            Point bp = blocks(rl);
            if (bp == null || bp.distanceTo(test)<Geometry.epsilon){
                double angle = Math.atan2(test.yValue() - spy, test.xValue() - spx) - axisAngle;
                if (angle < -Math.PI) angle += 2*Math.PI;
                if (angle > Math.PI) angle -= 2*Math.PI;
                if (angle < maxangle){
                    maxangle = angle;
                    res = test;
                }
            }
            i++;
        }
        return res;
    }
    
    public Point rightEdge(Line axis){
        double spy = axis.startPoint().yValue();
        double spx = axis.startPoint().xValue();
        double axisAngle = Math.atan2(axis.endPoint().yValue() - spy, axis.endPoint().xValue() - spx);
        int i = 0;
        Point res = null;
        double maxangle = -2.0*Math.PI;
        while(i < border.size()){
            Point test = (Point)(border.elementAt(i));
            Line rl = new Line(axis.startPoint(), test);
            Point bp = blocks(rl);
            if (bp == null || bp.distanceTo(test)<Geometry.epsilon){
                double angle = Math.atan2(test.yValue() - spy, test.xValue() - spx) - axisAngle;
                if (angle < -Math.PI) angle += 2*Math.PI;
                if (angle > Math.PI) angle -= 2*Math.PI;
                if (angle > maxangle){
                    maxangle = angle;
                    res = test;
                }
            }
            i++;
        }
        return res;
    }
    
    public void setLocation(Point p){
        Point cp = position();//((Point)(border.elementAt(0)));
        Point dp = p.sub(cp);
        setLocationRelative(dp);
    }
    
    public void setLocationRelative(Point dp){
        Vector nb = new Vector();
        for (int i = 0; i < border.size(); i++){
            Point c = (Point)(border.elementAt(i));
            nb.addElement(c.add(dp));
        }
        border = nb;
        setPosition(position().add(dp));
    }
    
    public void draw(Graphics g) {
        if (g == null) return;
        Polygon poly = new Polygon();
        for(int i = 0; i < border.size(); i++){ //populate the polygon
            Point p = (Point)(border.elementAt(i));
            poly.addPoint(Universe.u2sx(p), Universe.u2sy(p));
        }
        Color pc = g.getColor();
        g.setColor(color);
        g.fillPolygon(poly);
        g.setColor(pc);
    }
    
    public int type() {
        return Type.OBSTACLE;
    }
    
    /* #################################################### */
    
    public void showControl(){
        if (!ActorControlFrame.open.contains(this))
            new ActorControlFrame(this);
    }
    
    public String description(){
        String out;
        out = "Name: " + name() + "\n";
        out += "Type: Obstacle\n";
        out += "Location: "+position().toString();
        return out;
    }
    
    private JColorChooser cc;
    private JTextField nameField;
    
    public JPanel controlPanel() {
        //colour, radius, name, location
        JPanel p = new JPanel();
        cc = new JColorChooser(Color.red);
        //cc.setPreviewPanel(new JPanel());
        cc.getSelectionModel().addChangeListener(this);
        p.add(cc);
        JLabel nameLabel = new JLabel("Name: ");
        nameField = new JTextField(name(), 15);
        nameField.addActionListener(this);
        JPanel namePanel = new JPanel();
        namePanel.add(nameLabel);
        namePanel.add(nameField);
        cc.setPreviewPanel(namePanel);
        return p;
    }
    
    public void actionPerformed(java.awt.event.ActionEvent ae){
        if (ae.getSource().equals(nameField)){
            name = nameField.getText();
            nameField.setText(name);
        }
    }
    
    public void stateChanged(javax.swing.event.ChangeEvent e) {
        color = cc.getColor();
    }
    
    public abstract void resize(Point start, Point end);
    
}
