/*
 * Person.java
 *
 * Created on March 31, 2003, 6:06 PM
 */

package actor;

import java.util.*;
import java.awt.Graphics;
import java.awt.Color;
import geometry.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.Container;

/**
 *
 * @author  rod
 */
public class Person extends MovingActor implements GoalStackable, ChangeListener, ActionListener{
    static int personNumber = 0;
    static Color secondaryGoalColor = Color.lightGray;
    static double secondaryGoalRadius = 0.5; //metres
    double personalSpaceRadius;
    Vector knownPeople; //of PersonPair
    Stack goals;  //of GoalData
    Color color;
    static double personRadius = 0.2;
    Vector lines; //ditto
    Vector points; //yep
    
    /** Creates a new instance of Person */
    public Person(Point p, Vec initVel, Vec initAcc, Color c, double psr) {
        super(p, initVel, initAcc, 3, 5);
        knownPeople = new Vector();
        goals = new Stack();
        color = c;
        personalSpaceRadius = psr;
        personNumber++;
        setName("Person "+ personNumber);
        lines = new Vector();
        points = new Vector();
    }
    
    public double personalSpaceRadius(){
        return personalSpaceRadius;
    }
    
    public void resize(Point start, Point end){
        double d = end.distanceTo(start);
        personalSpaceRadius = d;
    }
    
    public void pushGoal(GoalStackable g, double importance){
        goals.push(new GoalData(g, true, importance, true));
    }
    
    public void addKnownPerson(Person p, double attraction){
        PersonPair pp = iKnowYou(p);
        if (pp == null){
            knownPeople.addElement(new PersonPair(p, attraction));
            return;
        }
        int index = knownPeople.indexOf(pp);
        knownPeople.setElementAt(new PersonPair(p, attraction), index);
    }
    
    private PersonPair iKnowYou(Person p){
        for (int i = 0; i < knownPeople.size(); i++){
            PersonPair pp = (PersonPair)(knownPeople.elementAt(i));
            if (pp.who().equals(p))
                return pp;
        }
        return null;
    }
    
    private boolean iHaveAGoal(){
        return (!(goals.size() == 0));
    }
    
    public GoalStackable currentGoal(){
        if (iHaveAGoal()){
            return ((GoalData)(goals.peek())).goal();
        }
        return null;
    }
    
    private boolean anyObstacleContains(Point p){
        for(int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Obstacle){
                if (((Obstacle)a).contains(p))
                    return true;
            }
        }
        return false;
    }
    
    private boolean iCanSeePrimaryGoal(){
        if (goals.isEmpty()) return false;
        Stack goalCopy=new Stack();
        GoalData gd = null;
        while(!goals.isEmpty()){
            gd = (GoalData)(goals.pop());
            goalCopy.push(gd);
            if (gd.isPrimary()) break;
        }
        while(!goalCopy.isEmpty()){
            goals.push(goalCopy.pop());
        }
        if (!gd.isPrimary()){
            return false;
        }
        GoalStackable cg = gd.goal();
        Line axis = new Line(position(),cg.position());
        for (int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Obstacle){
                if (((Obstacle)a).blocks(axis) != null){
                    return false;
                }
            }
        }
        return true;
    }
    
    public void removeSecondaryGoals(){
        if (goals.isEmpty()) return;
        while(!((GoalData)(goals.peek())).isPrimary()){
            Actor gs = (Actor)(((GoalData)(goals.pop())).goal());
            Universe.remove(gs);
        }
    }
    
    public void move(){
        if (Universe.paused){
            update();
            return;
        }
        if (iCanSeePrimaryGoal()){
            removeSecondaryGoals();
        }
        Line axis = null;
        lines = new Vector();
        points = new Vector();
        Vec goalDir = new Vec();
        if (iHaveAGoal()){
            /* NEGOTIATE OBSTACLES */
            boolean canGoAhead = false;
            while(!canGoAhead){
                GoalData gp = (GoalData)(goals.peek());
                GoalStackable currentGoal = gp.goal();
                double importance = gp.importance();
                Obstacle closest = null;
                double distToClosestObstacle2 = -1;
                axis = new Line(position(), currentGoal.position());
                double distToGoal2 = position().distanceTo2(currentGoal.position());
                lines.addElement(axis);
                Point pt = null;
                double distToObstacle2 = 0;
                for (int i = 0; i < Universe.actors.size(); i++){
                    Actor a = (Actor)(Universe.actors.elementAt(i));
                    if (a instanceof Obstacle){
                        if (((Obstacle)a).contains(position())){
                            setPosition(((Obstacle)a).closestPoint(position()));
                            velocity.scale(-1);
                            canGoAhead = true;
                            break;
                        }
                        pt = ((Obstacle)a).blocks(axis);
                        distToObstacle2 = distToGoal2 + 1;
                        if (pt != null){
                            distToObstacle2 = position().distanceTo2(pt);
                            if (distToObstacle2 < distToGoal2){
                                /*something's in the way*/
                                if((distToClosestObstacle2 == -1) || (distToObstacle2 < distToClosestObstacle2)){
                                    closest = (Obstacle)a;
                                    distToClosestObstacle2 = distToObstacle2;
                                    points.addElement(pt);
                                }
                            }
                        }
                    }
                }//end for (find closest obstacle)
                if (closest != null){ // gotta go 'round it
                    if (!gp.isPrimary()){ // can't be keeping track of all the intermediate points
                        gp.lostIt();
                    }
                    Point leftPoint = closest.leftEdge(axis);
                    Point rightPoint = closest.rightEdge(axis);
                    Line leftPath = new Line(position(), leftPoint);
                    Line rightPath = new Line(position(), rightPoint);
                    lines.addElement(leftPath);
                    lines.addElement(rightPath);
                    Line chosen;
                    double deviation = personalSpaceRadius*1.5;
                    if (gp.iKnowExactLocation()){
                        // minimize deviation from axis
                        if (Math.abs(axis.angle(leftPath)) < Math.abs(axis.angle(rightPath))){
                            chosen = leftPath;
                        }else{
                            chosen = rightPath;
                        }
                    }else{// i don't know, so minimize distance
                        if (position().distanceTo2(leftPoint) < position().distanceTo2(rightPoint)){
                            chosen = leftPath;
                        }else{
                            chosen = rightPath;
                        }
                    }
                    Line perp = chosen.perpendicular(chosen.endPoint());
                    lines.add(perp);
                    Point nt1 = perp.pointRelative1(chosen.endPoint(), deviation);
                    Point nt2 = perp.pointRelative2(chosen.endPoint(), deviation);
                    double dnt1 = closest.distanceToClosestPoint(nt1);
                    double dnt2 = closest.distanceToClosestPoint(nt2);
                    Point nt = null;
                    if ((dnt1 > dnt2) && !anyObstacleContains(nt1))
                        nt = nt1;
                    else
                        nt = nt2;
                    Goal subgoal = new Goal(secondaryGoalColor, secondaryGoalRadius, nt);
                    subgoal.setName(name()+"'s sub-goal #"+goals.size());
                    Universe.add(subgoal);
                    points.addElement(pt);
                    goals.push(new GoalData(subgoal, false, importance, true));
                }//else{ // there is no blocking object
                gp.foundIt();
                Vec p = new Vec(position());
                Vec p_hat = new Vec(currentGoal.position());
                Vec v_hat = p_hat.sub(p).scalarmul(Universe.k_p);
                Vec a_hat = v_hat.sub(velocity).scalarmul(Universe.k_v);
                goalDir = a_hat;
                canGoAhead = true;
                //}
            } // end going 'round it (while loop).
        }// end if (iHaveAGoal)
        else {
            goalDir = velocity;
            goalDir.scale(-0.5); // slow down if not headed anywhere
        }
        /* NAVIGATE AROUND / RESPOND TO OTHER PEOPLE */
        Vec peopleDir = new Vec();
        int interactions = 0;
        for (int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Person){
                Person p = (Person)a;
                PersonPair pp = iKnowYou(p);
                if (pp != null){ // I know this guy
                    Vec v = new Vec(new Line(position(), p.position()));
                    double dist = position().distanceTo(p.position());
                    double f = (-pp.attraction()*Universe.repulsionForce/dist);
                    v.setLength(f);
                    peopleDir=peopleDir.add(v);
                }
            }
        }//finished looking for people to respond to
        /* NOW MAINTAIN PERSONAL SPACE */
        Vec hardPersonalSpaceDir = new Vec();
        Vec softPersonalSpaceDir = new Vec();
        for (int i = 0; i < Universe.actors.size(); i++){
            Actor a = (Actor)(Universe.actors.elementAt(i));
            if (a instanceof Obstacle){
                Obstacle o = (Obstacle)a;
                double dist = o.distanceToClosestPoint(position());
                if (dist < personalSpaceRadius){
                    Point cp = o.closestPoint(position());
                    Line li = new Line(cp, position());
                    Vec ac = new Vec(li);
                    ac.scale(Universe.repulsionForce/dist/dist);
                    hardPersonalSpaceDir=hardPersonalSpaceDir.add(ac);
                }
            }// that's for obstacles
            else{
                if (!a.equals(this)){
                    if (goals.empty() || !a.equals(((GoalData)(goals.peek())).goal())){
                        if (!(a instanceof Goal)){ // goals aren't really physical
                            double dist = position().distanceTo(a.position());
                            if (dist < personalSpaceRadius){
                                Vec v = new Vec(new Line(a.position(), position()));
                                v.setLength(Universe.repulsionForce/dist/dist);
                                softPersonalSpaceDir=softPersonalSpaceDir.add(v);
                            }
                        }
                    }else{
                        double dist = position().distanceTo(a.position());
                        if (dist < personalSpaceRadius){
                            // trigger event?
                            GoalData gd = (GoalData)(goals.pop());
                            System.out.println(name()+" attained goal "+gd.goal().name()); // debug
                            if (!gd.isPrimary())
                                Universe.remove((Actor)(gd.goal()));
                        }
                    }
                }
            }
            
        }//personal space
        /* so, how to add these vectors...
         * probably goal, people, soft space, then hard space
         */
        Vec acc = new Vec();
        acc=acc.add(goalDir);
        acc=acc.add(peopleDir);
        acc=acc.add(softPersonalSpaceDir);
        acc.scale(1.0/3.0); // equal weights to these
        if (hardPersonalSpaceDir.length() > Geometry.epsilon)
            acc=hardPersonalSpaceDir;//### hmm. still doesn't guarantee...
        //acc.scale(10);
        acceleration = acc;
        update(); // call to super
    }
    
    public void draw(Graphics g) {
        if (g==null) return;
        int ulx = Universe.u2sx(position());
        int uly = Universe.u2sy(position());
        int r = Universe.u2s(personRadius);
        int psr = Universe.u2s(personalSpaceRadius);
        g.setColor(color);
        g.fillOval(ulx - r, uly - r, 2*r, 2*r);
        g.drawOval(ulx - psr, uly - psr, 2*psr, 2*psr);
        if (lines != null){
            for(int i=0; i<lines.size(); i++){
                Line l = (Line)(lines.elementAt(i));
                lines.removeElement(l);
                l.draw(g);
            }
        }
        if (points != null){
            for(int i=0; i<points.size(); i++){
                Point p = (Point)(points.elementAt(i));
                points.removeElement(p);
                p.draw(g);
            }
        }
        acceleration.draw(g, position(), Color.red);
        velocity.draw(g, position(), Color.green);
    }
    
    public int type() {
        return Type.PERSON;
    }
    
    /* #################################################### */
    
    public void showControl(){
        if (!ActorControlFrame.open.contains(this))
            new ActorControlFrame(this);
    }
    
    public String description(){
        String out;
        out = "Name: " + name() + "\n";
        out += "Type: Person\n";
        out += "Location: "+position().toString()+"\n";
        out += "Goals:\n";
        Stack g = new Stack();
        synchronized(goals){
            while(!goals.isEmpty()){
                GoalData gd =(GoalData)(goals.pop());
                out += "\t"+gd.goal().name()+"\n";
                g.push(gd);
            }
            while(!g.isEmpty()){
                GoalData gd =(GoalData)(g.pop());
                goals.push(gd);
            }
        }
        return out;
    }
    
    private JColorChooser cc;
    private JTextField nameField;
    private JTextField maxVelField;
    private JComboBox goalBox;
    private JComboBox peopleBox;
    
    public JPanel controlPanel() {
        //colour, radius, name, location
        JPanel p = new JPanel();
        cc = new JColorChooser(Color.red);
        //cc.setPreviewPanel(new JPanel());
        cc.getSelectionModel().addChangeListener(this);
        
        JPanel namePanel = new JPanel();
        
        JLabel nameLabel = new JLabel("Name: ");
        nameField = new JTextField(name(), 15);
        nameField.addActionListener(this);
        namePanel.add(nameLabel);
        namePanel.add(nameField);
        
        JLabel maxVelLabel = new JLabel("Max Speed: ");
        maxVelField = new JTextField(Double.toString(maxV), 5);
        maxVelField.addActionListener(this);
        namePanel.add(maxVelLabel);
        namePanel.add(maxVelField);
        
        JLabel goalLabel = new JLabel("New Goal: ");
        Vector goalVector = new Vector();
        synchronized(Universe.actors){
            for(int i = 0; i < Universe.actors.size(); i++){
                Actor a = (Actor)(Universe.actors.elementAt(i));
                if (a.canBeGoal() && !a.equals(this)){
                    goalVector.add(a);
                }
            }
        }
        if (!goalVector.isEmpty()){
            goalBox = new JComboBox(goalVector);
            goalBox.addActionListener(this);
            namePanel.add(goalLabel);
            namePanel.add(goalBox);
        }
        
        JLabel peopleLabel = new JLabel("New Known Person: ");
        Vector peopleVector = new Vector();
        synchronized(Universe.actors){
            for(int i = 0; i < Universe.actors.size(); i++){
                Actor a = (Actor)(Universe.actors.elementAt(i));
                if (a instanceof Person && !a.equals(this)){
                    peopleVector.add(a);
                }
            }
        }
        if (!peopleVector.isEmpty()){
            peopleBox = new JComboBox(peopleVector);
            peopleBox.addActionListener(this);
            namePanel.add(peopleLabel);
            namePanel.add(peopleBox);
        }
        
        cc.setPreviewPanel(namePanel);
        p.add(cc);
        return p;
    }
    
    public void actionPerformed(java.awt.event.ActionEvent ae){
        if (ae.getSource().equals(nameField)){
            name = nameField.getText();
            nameField.setText(name);
        }
        if (ae.getSource().equals(maxVelField)){
            maxV = (new Double(maxVelField.getText())).doubleValue();
            maxVelField.setText(Double.toString(maxV));
        }
        if(ae.getSource().equals(goalBox)){
            GoalStackable gs = (GoalStackable)(goalBox.getSelectedItem());
            GoalData gd = new GoalData(gs, true, 1.0, false);
            goals.push(gd);
            System.out.println(name()+" pushed new goal: "+gs.name());
            removeSecondaryGoals();
        }
        if(ae.getSource().equals(peopleBox)){
            Person p = (Person)(peopleBox.getSelectedItem());
            double att = Math.random()*2.0 - 1.0; // random like/dislike for now
            addKnownPerson(p, att);
            System.out.println(name()+" introduced new person: "+p.name()+" with attraction "+att);
        }
    }
    
    public void stateChanged(javax.swing.event.ChangeEvent e) {
        color = cc.getColor();
    }
    
    
    /* #################################################### */
    
    
    
    /* convenience classes for encapsulating knowledge about actors */
    class GoalData{
        private GoalStackable g;
        private boolean isPrimary;
        double importance; //0 to 1
        boolean iKnowExactLocation;
        
        public GoalData(GoalStackable goal, boolean isntSecondary, double imp, boolean iKnowExactLocation){
            g=goal;
            isPrimary=isntSecondary;
            importance = imp;
            this.iKnowExactLocation = iKnowExactLocation;
        }
        
        public boolean equals(GoalData gd){
            return gd.goal().equals(goal());
        }
        
        public boolean isPrimary(){
            return isPrimary;
        }
        
        public GoalStackable goal(){
            return g;
        }
        
        public double importance(){
            return importance;
        }
        
        public boolean iKnowExactLocation(){
            return iKnowExactLocation;
        }
        
        public void foundIt(){
            iKnowExactLocation = true;
        }
        
        public void lostIt(){
            iKnowExactLocation = false;
        }
    }
    
    class PersonPair{
        private Person p;
        private double attraction;  //let's say -1 is repulsed, 1 attracted
        
        public PersonPair(Person person, double attraction){
            p = person;
            this.attraction=attraction;
        }
        
        public double attraction(){
            return attraction;
        }
        
        public void setAttraction(double att){
            attraction = att;
        }
        
        public Person who(){
            return p;
        }
    }
}

