/*
 * Universe.java
 *
 * Created on April 7, 2003, 7:42 PM
 */

package actor;

import java.util.Vector;
import java.awt.Graphics;
import javax.swing.JPanel;
import javax.swing.JFrame;
import java.awt.Rectangle;
import geometry.Point;

import javax.swing.AbstractAction;
import javax.swing.Action;
import java.awt.event.ActionEvent;

/**
 *
 * @author  rod
 */
public class Universe{
    public static Vector actors;
    static int index;
    public static double scale;
    static JPanel jpanel; //where the Universe lives
    static double uminx, uminy, umaxx, umaxy; //Universe
    static int sminx, sminy, smaxx, smaxy; //Screen
    static boolean paused = true;
    static Point origin;
    static public double k_p = 2.0;
    static public double k_theta = 1.0;
    static public double k_v = 3.0;
    static public double k_omega = 1.0;
    static public double theta_turn = 0.03;
    static public double epsilon = 0.1;
    static public double repulsionForce = 5.0;
    boolean debug = true;
    static JFrame sf;
    
    /** Creates a new instance of Universe */
    public Universe(JPanel J, double scale, JFrame sf) { // scale is #pixels : 1 metre
        jpanel = J;
        actors = new Vector();
        Rectangle r = J.getBounds();
        sminx = 0;
        sminy = 0;
        smaxx = sminx+(int)(r.getMaxX());
        smaxy = sminy+(int)(r.getMaxY());
        uminx = (double)sminx/scale;
        uminy = (double)sminy/scale;
        umaxx = (double)smaxx/scale;
        umaxy = (double)smaxy/scale;
        origin = new Point(0,0);
        this.scale = scale;
        this.sf = sf;
    }
    
    public static void moveOrigin(Point udiff){
        Rectangle r = jpanel.getBounds();
        origin = origin.add(udiff);
        sminx = u2sx(udiff);
        sminy = u2sy(udiff);
        smaxx = sminx+(int)(r.getMaxX());
        smaxy = sminy+(int)(r.getMaxY());
        updatePanel();
    }
    
    public static void updatePanel(){
        if (jpanel==null) return; // sometimes...
        Rectangle r = jpanel.getBounds();
        if (r != null){ // i don't know why, but sometimes it's null...
            smaxx = sminx+(int)(r.getMaxX());
            smaxy = sminy+(int)(r.getMaxY());
            uminx = (double)sminx/scale;
            uminy = (double)sminy/scale;
            umaxx = (double)smaxx/scale;
            umaxy = (double)smaxy/scale;
        }
    }
    
    private  void printDimensions(){
        System.out.println("Screen: ("+sminx+","+sminy+") - ("+smaxx+","+smaxy+")");
        System.out.println("Virtual: ("+uminx+","+uminy+") - ("+umaxx+","+umaxy+")");
    }
    
    void changeScale(double newscale){
        scale = newscale;
        updatePanel();
    }
    
    public static boolean add(Actor a){
        return (actors.add(a));
    }
    
    public static void add(int i, Actor a){
        actors.add(i, a);
    }
    
    public Actor elementAt(int i){
        return (Actor)(actors.elementAt(i));
    }
    
    public int indexOf(Actor a){
        return actors.indexOf(a);
    }
    
    public static boolean remove(Actor a){
        return (actors.remove(a));
    }
    
    public static Actor remove(int i){
        return (Actor)(actors.remove(i));
    }
    
    public static Point centre(){
        //        return new Point ((uminx+umaxx)/2.0, (uminy+umaxy)/2.0);
        return s2u(jpanel.getWidth()/2, jpanel.getHeight()/2);
    }
    
    public static int u2sx(Point p){ //universal to screen x
        double dx = p.xValue() - uminx;
        double proportion = dx/(umaxx - uminx);
        double conv = proportion * (smaxx - sminx) + sminx;
        return sminx+(int)conv;
    }
    
    public static int u2sy(Point p){ //universal to screen y
        double dy = p.yValue() - uminy;
        double proportion = dy/(umaxy - uminy);
        double conv = proportion * (smaxy - sminy) + sminy;
        return sminy+(int)conv;
    }
    
    public static int u2s(double d){
        return (int)(d*scale);
    }
    
    public static Point s2u(int x, int y){ // screen (x,y) to universal
        int dx = x - sminx;
        double propx = (double)dx / (double)(smaxx-sminx);
        double ux = propx * (umaxx-uminx) + uminx;
        int dy = y - sminy;
        double propy = (double)dy / (double)(smaxy-sminy);
        double uy = propy * (umaxy-uminy) + uminy;
        return new Point(ux - sminx/scale, uy - sminy/scale);
    }
    
    public static void drawAll(Graphics g){
        //Graphics g = jpanel.getGraphics();
        synchronized(actors){
            if (g==null) return;
            if (actors == null) return;
            for(int i = 0; i < actors.size(); i++){
                if (actors == null) return;
                ((Actor)(actors.elementAt(i))).draw(g);
                Thread.yield();
            }
        }
    }
    
    //public static void moveAll(){
    //public static final Runnable moveAll = new Runnable(){
    //    public void run(){
    public static final Action moveAll = new AbstractAction(){
        public void actionPerformed(ActionEvent e){
            //while(true){
                synchronized(actors){
                    for(int i = 0; i < actors.size(); i++){
                        if (actors == null) return;
                        Actor a = (Actor)(actors.elementAt(i));
                        if(a instanceof MovingActor){
                            ((MovingActor)a).move();
                        }
                    }
                }
                sf.repaint();
                Thread.yield();
            //}
        }
    };
    
    public static void pauseUnpause(){
        paused = !paused;
    }
}