/*
 * Line.java
 *
 * Created on March 31, 2003, 6:08 PM
 */

package geometry;

import java.awt.Graphics;
import java.awt.Color;
import actor.Universe;
/**
 *
 * @author  rod
 */
public class Line {
    private Point start;
    private Point end;
    
    /** Creates a new instance of Line */
    public Line() {
        start = new Point();
        end = new Point();
    }
    
    public Line(Point start, Point end){
        this.start = start;
        this.end = end;
    }
    
    public Line(Point onPoint, double slope){
        start = onPoint;
        end = new Point(onPoint.xValue() + 1, onPoint.yValue() + slope);
    }
    
    public Point startPoint(){
        return start;
    }
    
    public Point endPoint(){
        return end;
    }
    
    public Line bisector(){
        Point p = new Point(start.yValue(), end.xValue());
        Point q = new Point(end.yValue(), start.xValue());
        return new Line(p, q);
    }
    
    public double length(){
        return start.distanceTo(end);
    }
    
    public double slope(){
        double dy = start.yValue() - end.yValue();
        double dx = start.xValue() - end.xValue();
        double s = 0;
        return (dy/dx);
    }
    
    public double yint(){
        double x = start.xValue();
        double y = start.yValue();
        y -= x * slope();
        return y;
    }
    
    public double yatx(double x){
        double y=0;
        y = slope() * x + yint();
        return y;
    }
    
    public Point intersect(Line l){
        double m = slope();
        double lm = l.slope();
        if (m == lm) return null;
        double ix = 0;
        double iy = 0;
        if (Double.isInfinite(m) || Double.isInfinite(lm)){
            if (Double.isInfinite(m)){
                ix = start.xValue();
                iy = l.yatx(ix);
                return new Point(ix,iy);
            }else{
                ix = l.startPoint().xValue();
                iy = yatx(ix);
                return new Point(ix,iy);
            }
        }
        ix = (l.yint() - yint()) / (slope() - l.slope());
        iy = yatx(ix);
        return new Point(ix, iy);
    }
    
    public Point intersectSegment(Line l){ //this is the segment, l is a line
        Point p = intersect(l);
        if (p == null) return null;
        if (contains(p)) return p;
        return null;
    }
    
    public Point intersectRay(Line l){ // this is the ray, l is a line segment
        /* need this to decide if we're inside a figure */
        //System.out.println("Hmm");
        Point i = l.intersectSegment(this);
        if (i == null) return null;
        Vec piv = new Vec(new Line(start, i));
        piv.normalize();
        Vec lv = new Vec(this);
        lv.normalize();
        if ((piv.sub(lv)).length() > Geometry.epsilon){
            i = null;
        }
        return i;
    }
    
    public Line parallel(Point p){
        Point q = new Point(p.xValue() + 1, p.yValue() + slope());
        return new Line(p, q);
    }
    
    public double distanceTo(Point p){
        Point inter = project(p);
        return p.distanceTo(inter);
    }
    
    public double distanceTo2(Point p){
        Point inter = project(p);
        return p.distanceTo2(inter);
    }
    
    public Line perpendicular(Point p){
        double m = slope();
        if (Double.isInfinite(m)){
            return new Line(p, new Point(p.xValue() + 1, p.yValue()));
        }
        if (m == 0){ // horizontal
            return new Line(p, new Point(p.xValue(), p.yValue() + 1));
        }
        m = -1.0/m;
        return new Line(p, m);
    }
    
    public boolean contains(Point p){ // is between start & end
        double l1 = p.distanceTo(start);
        double l2 = p.distanceTo(end);
        double length = start.distanceTo(end);
        return (Math.abs(l1 + l2 - length) < Geometry.epsilon);
    }
    
    /* angle counterclockwise from this to param */
    public double angle(Line l){
        Vec v1 = new Vec(this);
        Vec v2 = new Vec(l);
        return v1.angle(v2);
    }
    
    public Point project(Point p){
        Line perp = perpendicular(p);
        return intersect(perp);
    }
    
    public void draw(Graphics g){
        if (g==null) return;
        int bx = Universe.u2sx(start);
        int by = Universe.u2sy(start);
        int ex = Universe.u2sx(end);
        int ey = Universe.u2sy(end);
        Color pc = g.getColor();
        g.setColor(Color.gray);
        g.drawLine(bx,by,ex,ey);
        g.setColor(pc);
        //        System.out.println("dl: "+bx+" "+by+" "+ex+" "+ey);
    }
    
    public Point pointRelative1(Point p, double d) {
        double xp = p.xValue();
        double yp = p.yValue();
        double theta = Math.atan(slope());
        if(Double.isInfinite(theta)){
            return new Point(xp, yp-d);
        }
        return new Point(xp+d*Math.cos(theta), yp+d*Math.sin(theta));
    }
    
    public Point pointRelative2(Point p, double d) {
        double xp = p.xValue();
        double yp = p.yValue();
        double theta = Math.atan(slope());
        if(Double.isInfinite(theta)){
            return new Point(xp, yp+d);
        }
        return new Point(xp-d*Math.cos(theta), yp-d*Math.sin(theta));
    }
}
