/*
 * Vector.java
 *
 * Created on March 31, 2003, 6:18 PM
 */

package geometry;

import java.awt.Color;
import java.awt.Graphics;
import actor.Universe;

/**
 *
 * @author  rod
 */
public class Vec extends Point{
    
    /** Creates a new instance of Vector */
    public Vec() {
        super();
    }
    
    public Vec(double i, double j){
        super(i, j);
    }
    
    public Vec(Line l){
        super(l.endPoint().xValue() - l.startPoint().xValue(), 
        l.endPoint().yValue() - l.startPoint().yValue()); 
     /*   System.out.println(l.startPoint().toString()+" - "+l.endPoint().toString()); */
    }
    
    public Vec(Point p){
        super(p.xValue(), p.yValue());
    }
    
    public double dot(Vec p){
        double rx = x * p.xValue();
        double ry = y * p.yValue();
        return (rx + ry);
    }
    
    public double length(){
        return Math.sqrt(x*x + y*y);
    }
    
    public void setLength(double len){
        normalize();
        scale(len);
    }
    
    public void normalize(){
        scale(1/length());
    }
    
    public double angle(Vec v){
        double uvcosq = dot(v);
        uvcosq /= (length()*v.length());
        return Math.acos(uvcosq);
    }
    
    public boolean equals(Vec v){
        double dx = Math.abs(x-v.xValue());
        double dy = Math.abs(y-v.yValue());
        return ((dx < Geometry.epsilon) && (dy < Geometry.epsilon));
    }
    
    public Vec add(Vec v){
        return new Vec(x + v.xValue(), y + v.yValue());
    }
    
    public Vec sub(Vec v){
        return new Vec(x - v.xValue(), y - v.yValue());
    }
    
    public Vec scalarmul(double s){
        return new Vec(s*x, s*y);
    }

	public Vec reflect(Line l){
		Point endPoint = new Point(x,y);
		Vec reflectAxis = new Vec(l);
		Line lv = new Line(new Point(0,0), endPoint);
		Line xv = new Line(new Point(0,0), l.endPoint().sub(l.startPoint()));
		if (Double.isInfinite(xv.slope())){
			return new Vec(-x, y);
		}
		if (xv.slope() == 0){
			return new Vec(x, -y);
		}
		Point p = xv.project(endPoint);
		Line perp = xv.perpendicular(p);
		double dx = lv.endPoint().xValue() - p.xValue();
		Point rp = new Point(lv.endPoint().xValue()-2*dx,perp.yatx(lv.endPoint().xValue()-2*dx));
		return new Vec(rp);

	}
    
    public String toString(){
        return ("["+x+" "+y+"]");
    }
    
    public void draw(Graphics g, Point startPoint, Color c){
        int sx = Universe.u2sx(startPoint);
        int sy = Universe.u2sy(startPoint);
        int ex = sx + Universe.u2s(x);
        int ey = sy + Universe.u2s(y);
        g.setColor(c);
        g.drawLine(sx, sy, ex, ey);
        g.fillOval(ex-1,ey-1,2,2);
    }
    
}
